//
// File:       oclScan.c
//
// Abstract:   This example shows how to perform an efficient parallel prefix sum (aka Scan)
//             using OpenCL.  Scan is a common data parallel primitive which can be used for 
//             variety of different operations -- this example uses local memory for storing
//             partial sums and avoids memory bank conflicts on architectures which serialize
//             memory operations that are serviced on the same memory bank by offsetting the
//             loads and stores based on the size of the local group and the number of
//             memory banks (see appropriate macro definition).  As a result, this example
//             requires that the local group size > 1.
//
// Acknowledgement: 
//              This is a modified version of "OpenCL Parallel Prefix Sum (aka Scan) Example",
//              a sample code from  "Mac Developer Library".
//
////////////////////////////////////////////////////////////////////////////////////////////////////


#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <CL/cl.h>
#define KER_DIR "./kernels/"
////////////////////////////////////////////////////////////////////////////////////////////////////

int     outSize     = 0;
int     paraCount   = 0;
size_t* paraSizes   = 0;
void**  paras       = 0;

#define NUM_BANKS       (16)


#define min(A,B) ((A) < (B) ? (A) : (B))

////////////////////////////////////////////////////////////////////////////////////////////////////

cl_command_queue        ComputeCommands;
cl_context              ComputeContext;
cl_program              ComputeProgram;
cl_kernel*              ComputeKernels;
cl_mem*                 ScanPartialSums = 0;
unsigned int            ElementsAllocated = 0;
unsigned int            LevelsAllocated = 0;

////////////////////////////////////////////////////////////////////////////////////////////////////

enum KernelMethods
{
    PRESCAN                                         = 0,
    PRESCAN_STORE_SUM                               = 1,
    PRESCAN_STORE_SUM_NON_POWER_OF_TWO              = 2,
    PRESCAN_NON_POWER_OF_TWO                        = 3,
    UNIFORM_ADD                                     = 4,
    PRESCAN_PREPROCESS                              = 5,
    PRESCAN_STORE_SUM_PREPROCESS                    = 6,
    PRESCAN_STORE_SUM_NON_POWER_OF_TWO_PREPROCESS   = 7,
    PRESCAN_NON_POWER_OF_TWO_PREPROCESS             = 8
};

static const char* KernelNames[] =
{
    "PreScanKernel",
    "PreScanStoreSumKernel",
    "PreScanStoreSumNonPowerOfTwoKernel",
    "PreScanNonPowerOfTwoKernel",
    "UniformAddKernel",
    "PreScanKernel_Preprocess",
    "PreScanStoreSumKernel_Preprocess",
    "PreScanStoreSumNonPowerOfTwoKernel_Preprocess",
    "PreScanNonPowerOfTwoKernel_Preprocess"  
};

static const unsigned int KernelCount = sizeof(KernelNames) / sizeof(char *);

////////////////////////////////////////////////////////////////////////////////////////////////////

static char *
LoadProgramSourceFromFile(const char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

bool IsPowerOfTwo(int n)
{
    return ((n&(n-1))==0) ;
}

int floorPow2(int n)
{
    
    int exp;

    frexp((float)n, &exp);
   
    return 1 << (exp - 1);
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int 
CreatePartialSumBuffers(unsigned int group_size, unsigned int count)
{
    ElementsAllocated = count;

    unsigned int element_count = count;

    int level = 0;

    do
    {
        
        unsigned int group_count = (int)fmax(1, (int)ceil((float)element_count / (2.0f * group_size)));
        
        if (group_count > 1)
        {
            level++;
        }
        element_count = group_count;
        
    } while (element_count > 1);

    ScanPartialSums = (cl_mem*) malloc(level * sizeof(cl_mem));
    LevelsAllocated = level;
    memset(ScanPartialSums, 0, sizeof(cl_mem) * level);
    
    element_count = count;
    level = 0;
    
    do
    {       
        unsigned int group_count = (int)fmax(1, (int)ceil((float)element_count / (2.0f * group_size)));
        if (group_count > 1) 
        {
            size_t buffer_size = group_count * outSize;
            ScanPartialSums[level++] = clCreateBuffer(ComputeContext, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
        }

        element_count = group_count;

    } while (element_count > 1);

    return CL_SUCCESS;
}

void 
ReleasePartialSums(void)
{
    unsigned int i;
    for (i = 0; i < LevelsAllocated; i++)
    {
        clReleaseMemObject(ScanPartialSums[i]);
    }    
    
    free(ScanPartialSums);
    ScanPartialSums = 0;
    ElementsAllocated = 0;
    LevelsAllocated = 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int
PreScan(
    size_t *global, 
    size_t *local, 
    size_t shared, 
    cl_mem output_data, 
    cl_mem input_data, 
    unsigned int n,
    int group_index, 
    int base_index,
    int level
    )
{

    unsigned int k = (level==0)?PRESCAN_PREPROCESS:PRESCAN;    
    unsigned int a = 0;

    int err = CL_SUCCESS;
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &output_data);  
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &input_data);
    err |= clSetKernelArg(ComputeKernels[k],  a++, shared,         0);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &group_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &base_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &n);
    if(level == 0)
    {
        for(int i=0;i<paraCount;i++)
        {
            err |= clSetKernelArg(ComputeKernels[k],  a++, paraSizes[i], paras[i]);
        }      
    }
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to set kernel arguments!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    err = CL_SUCCESS;
    err |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernels[k], 1, NULL, global, local, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to execute kernel!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    return CL_SUCCESS;
}

int
PreScanStoreSum(
    size_t *global, 
    size_t *local, 
    size_t shared, 
    cl_mem output_data, 
    cl_mem input_data, 
    cl_mem partial_sums,
    unsigned int n,
    int group_index, 
    int base_index,
    int level
    )
{
    unsigned int k = (level==0)?PRESCAN_STORE_SUM_PREPROCESS:PRESCAN_STORE_SUM;
    unsigned int a = 0;

    int err = CL_SUCCESS;
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &output_data);  
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &input_data);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &partial_sums);
    err |= clSetKernelArg(ComputeKernels[k],  a++, shared,         0);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &group_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &base_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &n);
    if(level == 0)
    {
        for(int i=0;i<paraCount;i++)
        {
            err |= clSetKernelArg(ComputeKernels[k],  a++, paraSizes[i], paras[i]);
        }         
    }
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to set kernel arguments!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    err = CL_SUCCESS;
    err |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernels[k], 1, NULL, global, local, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to execute kernel!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }
    
    return CL_SUCCESS;
}

int
PreScanStoreSumNonPowerOfTwo(
    size_t *global, 
    size_t *local, 
    size_t shared, 
    cl_mem output_data, 
    cl_mem input_data, 
    cl_mem partial_sums,
    unsigned int n,
    int group_index, 
    int base_index,
    int level
    )
{
    unsigned int k = (level==0)? PRESCAN_STORE_SUM_NON_POWER_OF_TWO_PREPROCESS: PRESCAN_STORE_SUM_NON_POWER_OF_TWO;
    unsigned int a = 0;

    int err = CL_SUCCESS;
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &output_data);  
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &input_data);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &partial_sums);
    err |= clSetKernelArg(ComputeKernels[k],  a++, shared,         0);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &group_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &base_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &n);
    if(level == 0)
    {
        for(int i=0;i<paraCount;i++)
        {
            err |= clSetKernelArg(ComputeKernels[k],  a++, paraSizes[i], paras[i]);
        }          
    }
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to set kernel arguments!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    err = CL_SUCCESS;
    err |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernels[k], 1, NULL, global, local, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to execute kernel!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    return CL_SUCCESS;
}

int
PreScanNonPowerOfTwo(
    size_t *global, 
    size_t *local, 
    size_t shared, 
    cl_mem output_data, 
    cl_mem input_data, 
    unsigned int n,
    int group_index, 
    int base_index,
    int level
    )
{

    unsigned int k = (level==0)?PRESCAN_NON_POWER_OF_TWO_PREPROCESS:PRESCAN_NON_POWER_OF_TWO;
    unsigned int a = 0;

    int err = CL_SUCCESS;
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &output_data);  
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &input_data);
    err |= clSetKernelArg(ComputeKernels[k],  a++, shared,         0);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &group_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &base_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &n);
    if(level == 0)
    {
        for(int i=0;i<paraCount;i++)
        {
            err |= clSetKernelArg(ComputeKernels[k],  a++, paraSizes[i], paras[i]);
        }           
    }
      
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to set kernel arguments!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    err = CL_SUCCESS;
    err |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernels[k], 1, NULL, global, local, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to execute kernel!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }
    return CL_SUCCESS;
}

int
UniformAdd(
    size_t *global, 
    size_t *local, 
    cl_mem output_data, 
    cl_mem partial_sums, 
    unsigned int n, 
    unsigned int group_offset, 
    unsigned int base_index)
{
    unsigned int k = UNIFORM_ADD;
    unsigned int a = 0;

    int err = CL_SUCCESS;
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &output_data);  
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_mem), &partial_sums);
    err |= clSetKernelArg(ComputeKernels[k],  a++, outSize       ,  0);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &group_offset);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &base_index);
    err |= clSetKernelArg(ComputeKernels[k],  a++, sizeof(cl_int), &n);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to set kernel arguments!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    err = CL_SUCCESS;
    err |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernels[k], 1, NULL, global, local, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: %s: Failed to execute kernel!\n", KernelNames[k]);
        return EXIT_FAILURE;
    }

    return CL_SUCCESS;
}

int 
PreScanBufferRecursive(
    cl_mem output_data, 
    cl_mem input_data, 
    int group_size,
    int max_work_item_count,
    int element_count, 
    int level)
{
    unsigned int group_count = (int)fmax(1.0f, (int)ceil((float)element_count / (2.0f * group_size)));
    unsigned int work_item_count = 0;

    if (group_count > 1)
        work_item_count = group_size;
    else if (IsPowerOfTwo(element_count))
        work_item_count = element_count / 2;
    else
        work_item_count = floorPow2(element_count);
        
    work_item_count = (work_item_count > max_work_item_count) ? max_work_item_count : work_item_count;

    unsigned int element_count_per_group = work_item_count * 2;
    unsigned int last_group_element_count = element_count - (group_count-1) * element_count_per_group;
    unsigned int remaining_work_item_count = (int)fmax(1.0f, last_group_element_count / 2);
    remaining_work_item_count = (remaining_work_item_count > max_work_item_count) ? max_work_item_count : remaining_work_item_count;
    unsigned int remainder = 0;
    size_t last_shared = 0;

    
    if (last_group_element_count != element_count_per_group)
    {
        remainder = 1;

        if(!IsPowerOfTwo(last_group_element_count))
            remaining_work_item_count = floorPow2(last_group_element_count);    
        
        remaining_work_item_count = (remaining_work_item_count > max_work_item_count) ? max_work_item_count : remaining_work_item_count;
        unsigned int padding = (2 * remaining_work_item_count) / NUM_BANKS;
        last_shared = outSize * (2 * remaining_work_item_count + padding);
    }

    remaining_work_item_count = (remaining_work_item_count > max_work_item_count) ? max_work_item_count : remaining_work_item_count;
    size_t global[] = { (int)fmax(1, group_count - remainder) * work_item_count, 1 };
    size_t local[]  = { work_item_count, 1 };  

    unsigned int padding = element_count_per_group / NUM_BANKS;
    size_t shared = outSize * (element_count_per_group + padding);
    
    cl_mem partial_sums = ScanPartialSums[level];
    int err = CL_SUCCESS;
    
    if (group_count > 1)
    {
        err = PreScanStoreSum(global, local, shared, output_data, input_data, partial_sums, work_item_count * 2, 0, 0, level);
        if(err != CL_SUCCESS)
            return err;
            
        if (remainder)
        {
            size_t last_global[] = { 1 * remaining_work_item_count, 1 };
            size_t last_local[]  = { remaining_work_item_count, 1 };  

            err = PreScanStoreSumNonPowerOfTwo(
                    last_global, last_local, last_shared, 
                    output_data, input_data, partial_sums,
                    last_group_element_count, 
                    group_count - 1, 
                    element_count - last_group_element_count,
                    level
                    );    
        
            if(err != CL_SUCCESS)
                return err;			
			
        }

        err = PreScanBufferRecursive(partial_sums, partial_sums, group_size, max_work_item_count, group_count, level + 1);
        if(err != CL_SUCCESS)
            return err;
            
        err = UniformAdd(global, local, output_data, partial_sums,  element_count - last_group_element_count, 0, 0);
        if(err != CL_SUCCESS)
            return err;
        
        if (remainder)
        {
            size_t last_global[] = { 1 * remaining_work_item_count, 1 };
            size_t last_local[]  = { remaining_work_item_count, 1 };  

            err = UniformAdd(
                    last_global, last_local, 
                    output_data, partial_sums,
                    last_group_element_count, 
                    group_count - 1, 
                    element_count - last_group_element_count);
                
            if(err != CL_SUCCESS)
                return err;
        }
    }
    else if (IsPowerOfTwo(element_count))
    {
        err = PreScan(global, local, shared, output_data, input_data, work_item_count * 2, 0, 0, level);
        if(err != CL_SUCCESS)
            return err;
    }
    else
    {
        err = PreScanNonPowerOfTwo(global, local, shared, output_data, input_data, element_count, 0, 0, level);
        if(err != CL_SUCCESS)
            return err;
    }

    return CL_SUCCESS;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 
typedef struct
{
    cl_context ComputeContext;
    cl_command_queue ComputeCommands;
    cl_program ComputeProgram;
    cl_kernel* ComputeKernels;
    unsigned int workgroupSize;
    unsigned int maxWorkItemsCount;
    size_t outSize;
    int paraCount;
    size_t* paraSizes;    
} oclScanner;

int oclScan( 
    oclScanner* scanner, 
    cl_mem output_data,  
    cl_mem input_data, 
    unsigned int count, 
    void** _paras)
{
   
    ComputeContext = scanner->ComputeContext;   
    ComputeCommands = scanner->ComputeCommands;
    ComputeProgram = scanner->ComputeProgram;    
    ComputeKernels = scanner->ComputeKernels;
    outSize = scanner->outSize;
    paraCount = scanner->paraCount;
    paraSizes = scanner->paraSizes;
    paras = _paras;
    
    CreatePartialSumBuffers(scanner->workgroupSize, count);
    
    PreScanBufferRecursive(output_data, input_data, scanner->workgroupSize, scanner->maxWorkItemsCount, count, 0);
    
    ReleasePartialSums();  
}

oclScanner* oclCreateScanner(
    cl_context context, 
    cl_command_queue queue, 
    cl_device_id device,
    unsigned int workgroupSize,
    unsigned int maxWorkItemsCount, 
    const char* presource_filename, 
    size_t outSize, 
    int paraCount,
    size_t* paraSizes,
    int* err)
{
    oclScanner* scanner = (oclScanner*) malloc(sizeof(oclScanner));
    scanner->ComputeContext = context;
    scanner->ComputeCommands = queue;
    scanner->ComputeProgram = 0;
    scanner->ComputeKernels = 0;
    scanner->workgroupSize = workgroupSize;
    scanner->maxWorkItemsCount = maxWorkItemsCount;
    scanner->outSize = outSize;
    scanner->paraCount = paraCount;
    scanner->paraSizes = paraSizes;    
    
    int i = 0;
    // Load the compute program from disk into a cstring buffer
    //
    const char* mainsource_filename = KER_DIR"scan_kernel.cl";
    char *mainsource = LoadProgramSourceFromFile(mainsource_filename);
    char *presource = LoadProgramSourceFromFile(presource_filename);
    if(!mainsource||!presource)
    {
        printf("Error: Failed to load compute program from file!\n");
        *err = EXIT_FAILURE;
        return 0;  
    }
    
    char* source = (char*) malloc(sizeof(char) *(strlen(mainsource)+strlen(presource)+1));
    strcat(strcpy(source, presource), mainsource);
    free(mainsource);
    free(presource);
    
    // Create the compute program from the source buffer
    //
    scanner->ComputeProgram = clCreateProgramWithSource(scanner->ComputeContext, 1, (const char **) & source, NULL, err);
    if (!(scanner->ComputeProgram) || *err != CL_SUCCESS)
    {
        printf("%s\n", source);
        printf("Error: Failed to create compute program!\n");
        return 0;
    }

    // Build the program executable
    //
    *err = clBuildProgram((scanner->ComputeProgram), 1,  &device, NULL, NULL, NULL);         
    if (*err != CL_SUCCESS)
    {
        size_t length;
        char build_log[2048];
        printf("%s\n", source);
        printf("Error: Failed to build program executable!\n");
        clGetProgramBuildInfo(scanner->ComputeProgram, device, CL_PROGRAM_BUILD_LOG, sizeof(build_log), build_log, &length);
        printf("%s\n", build_log);
        return 0;
    }

    scanner->ComputeKernels = (cl_kernel*) malloc(KernelCount * sizeof(cl_kernel));
    for(i = 0; i < KernelCount; i++)
    {    
        // Create each compute kernel from within the program
        //
        scanner->ComputeKernels[i] = clCreateKernel(scanner->ComputeProgram, KernelNames[i], err);
        if (!scanner->ComputeKernels[i] || *err != CL_SUCCESS)
        {
            printf("Error: Failed to create compute kernel!\n");
            return 0;
        }	
    }

    free(source);
    
    *err = CL_SUCCESS;

    return scanner;
}

int oclReleaseScanner(oclScanner* scanner)
{
    int i;
    for(i = 0; i < KernelCount; i++)
    {
        clReleaseKernel(scanner->ComputeKernels[i]);
    }

    clReleaseProgram(scanner->ComputeProgram);
    free(scanner->ComputeKernels);
    free(scanner);
}


